@@ -277,6 +277,70 @@ memcpy_to_tiled_x__swizzle_0(const void *src, void *dst, int bpp,
 	}
 }
 
+static fast_memcpy void
+memcpy_from_tiled_x__swizzle_0(const void *src, void *dst, int bpp,
+			       int32_t src_stride, int32_t dst_stride,
+			       int16_t src_x, int16_t src_y,
+			       int16_t dst_x, int16_t dst_y,
+			       uint16_t width, uint16_t height)
+{
+	const unsigned tile_width = 512;
+	const unsigned tile_height = 8;
+	const unsigned tile_size = 4096;
+
+	const unsigned cpp = bpp / 8;
+	const unsigned stride_tiles = src_stride / tile_width;
+	const unsigned swizzle_pixels = tile_width / cpp;
+	const unsigned tile_pixels = ffs(tile_width / cpp) - 1;
+	const unsigned tile_mask = (1 << tile_pixels) - 1;
+
+	unsigned x, y;
+
+	DBG(("%s(bpp=%d): src=(%d, %d), dst=(%d, %d), size=%dx%d, pitch=%d/%d\n",
+	     __FUNCTION__, bpp, src_x, src_y, dst_x, dst_y, width, height, src_stride, dst_stride));
+
+	dst = (uint8_t *)dst + dst_y * dst_stride + dst_x * cpp;
+
+	for (y = 0; y < height; ++y) {
+		const uint32_t sy = y + src_y;
+		const uint32_t tile_row =
+			(sy / tile_height * stride_tiles * tile_size +
+			 (sy & (tile_height-1)) * tile_width);
+		uint8_t *dst_row = (uint8_t *)dst + dst_stride * y;
+		uint32_t sx = src_x, offset;
+
+		x = width * cpp;
+		if (sx & (swizzle_pixels - 1)) {
+			const uint32_t swizzle_bound_pixels = ALIGN(sx + 1, swizzle_pixels);
+			const uint32_t length = min(src_x + width, swizzle_bound_pixels) - sx;
+			offset = tile_row +
+				(sx >> tile_pixels) * tile_size +
+				(sx & tile_mask) * cpp;
+			memcpy(dst_row, (const char *)src + offset, length * cpp);
+
+			dst_row += length * cpp;
+			x -= length * cpp;
+			sx += length;
+		}
+		while (x >= 512) {
+			assert((sx & tile_mask) == 0);
+			offset = tile_row + (sx >> tile_pixels) * tile_size;
+
+			memcpy(dst_row, (const char *)src + offset, 512);
+
+			dst_row += 512;
+			x -= 512;
+			sx += swizzle_pixels;
+		}
+		if (x) {
+			offset = tile_row +
+				(sx >> tile_pixels) * tile_size +
+				(sx & tile_mask) * cpp;
+			memcpy(dst_row, (const char *)src + offset, x);
+		}
+	}
+}
+
 fast_memcpy static void
 memcpy_to_tiled_x__swizzle_9(const void *src, void *dst, int bpp,
 			     int32_t src_stride, int32_t dst_stride,
@@ -346,6 +410,75 @@ memcpy_to_tiled_x__swizzle_9(const void *src, void *dst, int bpp,
 	}
 }
 
+fast_memcpy static void
+memcpy_from_tiled_x__swizzle_9(const void *src, void *dst, int bpp,
+			       int32_t src_stride, int32_t dst_stride,
+			       int16_t src_x, int16_t src_y,
+			       int16_t dst_x, int16_t dst_y,
+			       uint16_t width, uint16_t height)
+{
+	const unsigned tile_width = 512;
+	const unsigned tile_height = 8;
+	const unsigned tile_size = 4096;
+
+	const unsigned cpp = bpp / 8;
+	const unsigned stride_tiles = src_stride / tile_width;
+	const unsigned swizzle_pixels = 64 / cpp;
+	const unsigned tile_pixels = ffs(tile_width / cpp) - 1;
+	const unsigned tile_mask = (1 << tile_pixels) - 1;
+
+	unsigned x, y;
+
+	DBG(("%s(bpp=%d): src=(%d, %d), dst=(%d, %d), size=%dx%d, pitch=%d/%d\n",
+	     __FUNCTION__, bpp, src_x, src_y, dst_x, dst_y, width, height, src_stride, dst_stride));
+
+	dst = (uint8_t *)dst + dst_y * dst_stride + dst_x * cpp;
+
+	for (y = 0; y < height; ++y) {
+		const uint32_t sy = y + src_y;
+		const uint32_t tile_row =
+			(sy / tile_height * stride_tiles * tile_size +
+			 (sy & (tile_height-1)) * tile_width);
+		uint8_t *dst_row = (uint8_t *)dst + dst_stride * y;
+		uint32_t sx = src_x, offset;
+
+		x = width * cpp;
+		if (sx & (swizzle_pixels - 1)) {
+			const uint32_t swizzle_bound_pixels = ALIGN(sx + 1, swizzle_pixels);
+			const uint32_t length = min(src_x + width, swizzle_bound_pixels) - sx;
+			offset = tile_row +
+				(sx >> tile_pixels) * tile_size +
+				(sx & tile_mask) * cpp;
+			offset ^= (offset >> 3) & 64;
+
+			memcpy(dst_row, (const char *)src + offset, length * cpp);
+
+			dst_row += length * cpp;
+			x -= length * cpp;
+			sx += length;
+		}
+		while (x >= 64) {
+			offset = tile_row +
+				(sx >> tile_pixels) * tile_size +
+				(sx & tile_mask) * cpp;
+			offset ^= (offset >> 3) & 64;
+
+			memcpy(dst_row, (const char *)src + offset, 64);
+
+			dst_row += 64;
+			x -= 64;
+			sx += swizzle_pixels;
+		}
+		if (x) {
+			offset = tile_row +
+				(sx >> tile_pixels) * tile_size +
+				(sx & tile_mask) * cpp;
+			offset ^= (offset >> 3) & 64;
+			memcpy(dst_row, (const char *)src + offset, x);
+		}
+	}
+}
+
 fast_memcpy static void
 memcpy_to_tiled_x__swizzle_9_10(const void *src, void *dst, int bpp,
 				int32_t src_stride, int32_t dst_stride,
@@ -415,6 +548,75 @@ memcpy_to_tiled_x__swizzle_9_10(const void *src, void *dst, int bpp,
 	}
 }
 
+fast_memcpy static void
+memcpy_from_tiled_x__swizzle_9_10(const void *src, void *dst, int bpp,
+				  int32_t src_stride, int32_t dst_stride,
+				  int16_t src_x, int16_t src_y,
+				  int16_t dst_x, int16_t dst_y,
+				  uint16_t width, uint16_t height)
+{
+	const unsigned tile_width = 512;
+	const unsigned tile_height = 8;
+	const unsigned tile_size = 4096;
+
+	const unsigned cpp = bpp / 8;
+	const unsigned stride_tiles = src_stride / tile_width;
+	const unsigned swizzle_pixels = 64 / cpp;
+	const unsigned tile_pixels = ffs(tile_width / cpp) - 1;
+	const unsigned tile_mask = (1 << tile_pixels) - 1;
+
+	unsigned x, y;
+
+	DBG(("%s(bpp=%d): src=(%d, %d), dst=(%d, %d), size=%dx%d, pitch=%d/%d\n",
+	     __FUNCTION__, bpp, src_x, src_y, dst_x, dst_y, width, height, src_stride, dst_stride));
+
+	dst = (uint8_t *)dst + dst_y * dst_stride + dst_x * cpp;
+
+	for (y = 0; y < height; ++y) {
+		const uint32_t sy = y + src_y;
+		const uint32_t tile_row =
+			(sy / tile_height * stride_tiles * tile_size +
+			 (sy & (tile_height-1)) * tile_width);
+		uint8_t *dst_row = (uint8_t *)dst + dst_stride * y;
+		uint32_t sx = src_x, offset;
+
+		x = width * cpp;
+		if (sx & (swizzle_pixels - 1)) {
+			const uint32_t swizzle_bound_pixels = ALIGN(sx + 1, swizzle_pixels);
+			const uint32_t length = min(src_x + width, swizzle_bound_pixels) - sx;
+			offset = tile_row +
+				(sx >> tile_pixels) * tile_size +
+				(sx & tile_mask) * cpp;
+			offset ^= ((offset ^ (offset >> 1)) >> 3) & 64;
+
+			memcpy(dst_row, (const char *)src + offset, length * cpp);
+
+			dst_row += length * cpp;
+			x -= length * cpp;
+			sx += length;
+		}
+		while (x >= 64) {
+			offset = tile_row +
+				(sx >> tile_pixels) * tile_size +
+				(sx & tile_mask) * cpp;
+			offset ^= ((offset ^ (offset >> 1)) >> 3) & 64;
+
+			memcpy(dst_row, (const char *)src + offset, 64);
+
+			dst_row += 64;
+			x -= 64;
+			sx += swizzle_pixels;
+		}
+		if (x) {
+			offset = tile_row +
+				(sx >> tile_pixels) * tile_size +
+				(sx & tile_mask) * cpp;
+			offset ^= ((offset ^ (offset >> 1)) >> 3) & 64;
+			memcpy(dst_row, (const char *)src + offset, x);
+		}
+	}
+}
+
 fast_memcpy static void
 memcpy_to_tiled_x__swizzle_9_11(const void *src, void *dst, int bpp,
 				int32_t src_stride, int32_t dst_stride,
@@ -483,21 +685,99 @@ memcpy_to_tiled_x__swizzle_9_11(const void *src, void *dst, int bpp,
 	}
 }
 
-void choose_memcpy_to_tiled_x(struct kgem *kgem, int swizzling)
+fast_memcpy static void
+memcpy_from_tiled_x__swizzle_9_11(const void *src, void *dst, int bpp,
+				  int32_t src_stride, int32_t dst_stride,
+				  int16_t src_x, int16_t src_y,
+				  int16_t dst_x, int16_t dst_y,
+				  uint16_t width, uint16_t height)
+{
+	const unsigned tile_width = 512;
+	const unsigned tile_height = 8;
+	const unsigned tile_size = 4096;
+
+	const unsigned cpp = bpp / 8;
+	const unsigned stride_tiles = src_stride / tile_width;
+	const unsigned swizzle_pixels = 64 / cpp;
+	const unsigned tile_pixels = ffs(tile_width / cpp) - 1;
+	const unsigned tile_mask = (1 << tile_pixels) - 1;
+
+	unsigned x, y;
+
+	DBG(("%s(bpp=%d): src=(%d, %d), dst=(%d, %d), size=%dx%d, pitch=%d/%d\n",
+	     __FUNCTION__, bpp, src_x, src_y, dst_x, dst_y, width, height, src_stride, dst_stride));
+
+	dst = (uint8_t *)dst + dst_y * dst_stride + dst_x * cpp;
+
+	for (y = 0; y < height; ++y) {
+		const uint32_t sy = y + src_y;
+		const uint32_t tile_row =
+			(sy / tile_height * stride_tiles * tile_size +
+			 (sy & (tile_height-1)) * tile_width);
+		uint8_t *dst_row = (uint8_t *)dst + dst_stride * y;
+		uint32_t sx = src_x, offset;
+
+		x = width * cpp;
+		if (sx & (swizzle_pixels - 1)) {
+			const uint32_t swizzle_bound_pixels = ALIGN(sx + 1, swizzle_pixels);
+			const uint32_t length = min(src_x + width, swizzle_bound_pixels) - sx;
+			offset = tile_row +
+				(sx >> tile_pixels) * tile_size +
+				(sx & tile_mask) * cpp;
+			offset ^= ((offset ^ (offset >> 2)) >> 3) & 64;
+			memcpy(dst_row, (const char *)src + offset, length * cpp);
+
+			dst_row += length * cpp;
+			x -= length * cpp;
+			sx += length;
+		}
+		while (x >= 64) {
+			offset = tile_row +
+				(sx >> tile_pixels) * tile_size +
+				(sx & tile_mask) * cpp;
+			offset ^= ((offset ^ (offset >> 2)) >> 3) & 64;
+
+			memcpy(dst_row, (const char *)src + offset, 64);
+
+			dst_row += 64;
+			x -= 64;
+			sx += swizzle_pixels;
+		}
+		if (x) {
+			offset = tile_row +
+				(sx >> tile_pixels) * tile_size +
+				(sx & tile_mask) * cpp;
+			offset ^= ((offset ^ (offset >> 2)) >> 3) & 64;
+			memcpy(dst_row, (const char *)src + offset, x);
+		}
+	}
+}
+
+void choose_memcpy_tiled_x(struct kgem *kgem, int swizzling)
 {
 	switch (swizzling) {
 	default:
+		DBG(("%s: unknown swizzling, %d\n", __FUNCTION__, swizzling));
+		break;
 	case I915_BIT_6_SWIZZLE_NONE:
+		DBG(("%s: no swizzling\n", __FUNCTION__));
 		kgem->memcpy_to_tiled_x = memcpy_to_tiled_x__swizzle_0;
+		kgem->memcpy_from_tiled_x = memcpy_from_tiled_x__swizzle_0;
 		break;
 	case I915_BIT_6_SWIZZLE_9:
+		DBG(("%s: 6^9 swizzling\n", __FUNCTION__));
 		kgem->memcpy_to_tiled_x = memcpy_to_tiled_x__swizzle_9;
+		kgem->memcpy_from_tiled_x = memcpy_from_tiled_x__swizzle_9;
 		break;
 	case I915_BIT_6_SWIZZLE_9_10:
+		DBG(("%s: 6^9^10 swizzling\n", __FUNCTION__));
 		kgem->memcpy_to_tiled_x = memcpy_to_tiled_x__swizzle_9_10;
+		kgem->memcpy_from_tiled_x = memcpy_from_tiled_x__swizzle_9_10;
 		break;
 	case I915_BIT_6_SWIZZLE_9_11:
+		DBG(("%s: 6^9^11 swizzling\n", __FUNCTION__));
 		kgem->memcpy_to_tiled_x = memcpy_to_tiled_x__swizzle_9_11;
+		kgem->memcpy_from_tiled_x = memcpy_from_tiled_x__swizzle_9_11;
 		break;
 	}
 }
